######################################################
# Windrop  Vladislav 
# : 2.7.6
# : HTTP   GET, POST  HEAD,        Safesock.
######################################################

package provide http 2.7.6

namespace eval http {
variable http
if {![info exists http]} {
array set http {
	    -accept */*
	    -proxyhost {}
	    -proxyport {}
	    -proxyfilter http::ProxyRequired
	    -urlencoding utf-8
}
set http(-useragent) "Tcl http client package [package provide http]"
}

proc init {} {
for {set i 0} {$i <= 256} {incr i} {
set c [format %c $i]
if {![string match {[-._~a-zA-Z0-9]} $c]} {set map($c) %[format %.2x $i]}
}
set map(\n) %0d%0a
variable formMap [array get map]
variable socketmap
if {[info exists socketmap]} {
foreach {url sock} [array get socketmap] {
catch {close $sock}
}
}
array set socketmap {}
}
init
variable urlTypes
if {![info exists urlTypes]} {set urlTypes(http) [list 80 ::socket]}
variable encodings [string tolower [encoding names]]
variable defaultCharset
if {![info exists defaultCharset]} {set defaultCharset "iso8859-1"}
variable strict
if {![info exists strict]} {set strict 1}
variable defaultKeepalive
if {![info exists defaultKeepalive]} {set defaultKeepalive 0}
namespace export geturl config reset wait formatQuery register unregister
}

proc http::Log {args} {}

proc http::register {proto port command} {
variable urlTypes
set urlTypes($proto) [list $port $command]
}

proc http::unregister {proto} {
variable urlTypes
if {![info exists urlTypes($proto)]} {return -code error "unsupported url type \"$proto\""}
set old $urlTypes($proto)
unset urlTypes($proto)
return $old
}

proc http::config {args} {
variable http
set options [lsort [array names http -*]]
set usage [join $options ", "]
if {[llength $args] == 0} {
set result {}
foreach name $options {lappend result $name $http($name)}
return $result
}
set options [string map {- ""} $options]
set pat ^-(?:[join $options |])$
if {[llength $args] == 1} {
set flag [lindex $args 0]
if {![regexp -- $pat $flag]} {return -code error "Unknown option $flag, must be: $usage"}
return $http($flag)
} else {
foreach {flag value} $args {
if {![regexp -- $pat $flag]} {return -code error "Unknown option $flag, must be: $usage"}
set http($flag) $value
}
}
}

proc http::Finish {token {errormsg ""} {skipCB 0}} {
variable $token
upvar 0 $token state
global errorInfo errorCode
if {$errormsg ne ""} {
set state(error) [list $errormsg $errorInfo $errorCode]
set state(status) "error"
}
if {($state(status) eq "timeout") || ($state(status) eq "error") || ([info exists state(connection)] && ($state(connection) eq "close"))} then {CloseSocket $state(sock) $token}
if {[info exists state(after)]} {after cancel $state(after)}
if {[info exists state(-command)] && !$skipCB} {
if {[catch {eval $state(-command) {$token}} err]} {
if {$errormsg eq ""} {
set state(error) [list $err $errorInfo $errorCode]
set state(status) error
}
}
unset -nocomplain state(-command)
}
}

# http::CloseSocket -
#
#	Close a socket and remove it from the persistent sockets table.  If
#	possible an http token is included here but when we are called from a
#	fileevent on remote closure we need to find the correct entry - hence
#	the second section.

proc ::http::CloseSocket {s {token {}}} {
variable socketmap
catch {fileevent $s readable {}}
set conn_id {}
if {$token ne ""} {
variable $token
upvar 0 $token state
if {[info exists state(socketinfo)]} {set conn_id $state(socketinfo)}
} else {
set map [array get socketmap]
set ndx [lsearch -exact $map $s]
if {$ndx != -1} {
incr ndx -1
set conn_id [lindex $map $ndx]
}
}
if {$conn_id eq {} || ![info exists socketmap($conn_id)]} {
Log "Closing socket $s (no connection info)"
if {[catch {close $s} err]} {Log "Error: $err"}
} else {
if {[info exists socketmap($conn_id)]} {
 Log "Closing connection $conn_id (sock $socketmap($conn_id))"
if {[catch {close $socketmap($conn_id)} err]} {
Log "Error: $err"
}
unset socketmap($conn_id)
} else {Log "Cannot close connection $conn_id - no socket in socket map"}
}
}

# http::reset --
#
#	See documentation for details.

proc http::reset {token {why reset}} {
variable $token
upvar 0 $token state
set state(status) $why
catch {fileevent $state(sock) readable {}}
catch {fileevent $state(sock) writable {}}
Finish $token
if {[info exists state(error)]} {
set errorlist $state(error)
unset state
eval ::error $errorlist
}
}

# http::geturl --
#
#	Establishes a connection to a remote url via http.

proc http::geturl {url args} {
variable http
variable urlTypes
variable defaultCharset
variable defaultKeepalive
variable strict
if {![info exists http(uid)]} {set http(uid) 0}
set token [namespace current]::[incr http(uid)]
variable $token
upvar 0 $token state
reset $token
array set state {
	-binary		false
	-blocksize         8192
	-queryblocksize 8192
	-validate	        0
	-headers	        {}
	-timeout	        0
	-type		application/x-www-form-urlencoded
	-queryprogress	{}
	-protocol	        1.1
	binary		0
	state		        connecting
	meta		        {}
	coding		{}
	currentsize	0
	totalsize	        0
	querylength	0
	queryoffset	0
	type		        text/html
	body		        {}
	status		""
	http		        ""
	connection	close
}
set state(-keepalive) $defaultKeepalive
set state(-strict) $strict
array set type {
	-binary		boolean
	-blocksize	integer
	-queryblocksize integer
	-strict		boolean
	-timeout	integer
	-validate	boolean
}
set state(charset)	$defaultCharset
set options {
	-binary -blocksize -channel -command -handler -headers -keepalive
	-method -myaddr -progress -protocol -query -queryblocksize
	-querychannel -queryprogress -strict -timeout -type -validate
}
set usage [join [lsort $options] ", "]
set options [string map {- ""} $options]
set pat ^-(?:[join $options |])$
foreach {flag value} $args {
if {[regexp -- $pat $flag]} {
if {[info exists type($flag)] && ![string is $type($flag) -strict $value]} then {
unset $token
return -code error "Bad value for $flag ($value), must be $type($flag)"
}
set state($flag) $value
} else {
unset $token
return -code error "Unknown option $flag, can be: $usage"
}
}
set isQueryChannel [info exists state(-querychannel)]
set isQuery [info exists state(-query)]
if {$isQuery && $isQueryChannel} {
unset $token
return -code error "Can't combine -query and -querychannel options!"
}
set URLmatcher {(?x)		# this is _expanded_ syntax
	^
	(?: (\w+) : ) ?			# <protocol scheme>
	(?: //
	    (?:
		(
		    [^@/\#?]+		# <userinfo part of authority>
		) @
	    )?
	    ( [^/:\#?]+ )		# <host part of authority>
	    (?: : (\d+) )?		# <port part of authority>
	)?
	( / [^\#]*)?			# <path> (including query)
	(?: \# (.*) )?			# <fragment>
	$
}
if {![regexp -- $URLmatcher $url -> proto user host port srvurl]} {
unset $token
return -code error "Unsupported URL: $url"
}
if {$host eq ""} {
unset $token
return -code error "Missing host part: $url"
}
if {$port ne "" && $port > 65535} {
unset $token
return -code error "Invalid port number: $port"
}
if {$user ne ""} {
set validityRE {(?xi)
	    ^
	    (?: [-\w.~!$&'()*+,;=:] | %[0-9a-f][0-9a-f] )+
	    $
}
if {$state(-strict) && ![regexp -- $validityRE $user]} {
unset $token
if {[regexp {(?i)%(?![0-9a-f][0-9a-f]).?.?} $user bad]} {
return -code error "Illegal encoding character usage \"$bad\" in URL user"
}
return -code error "Illegal characters in URL user"
}
}
if {$srvurl ne ""} {
set validityRE {(?xi)
	    ^
	    # Path part (already must start with / character)
	    (?:	      [-\w.~!$&'()*+,;=:@/]  | %[0-9a-f][0-9a-f] )*
	    # Query part (optional, permits ? characters)
	    (?: \? (?: [-\w.~!$&'()*+,;=:@/?] | %[0-9a-f][0-9a-f] )* )?
	    $
}
if {$state(-strict) && ![regexp -- $validityRE $srvurl]} {
unset $token
if {[regexp {(?i)%(?![0-9a-f][0-9a-f])..} $srvurl bad]} {
return -code error "Illegal encoding character usage \"$bad\" in URL path"
}
return -code error "Illegal characters in URL path"
}
} else {
set srvurl /
}
if {$proto eq ""} {set proto http}
if {![info exists urlTypes($proto)]} {
unset $token
return -code error "Unsupported URL type \"$proto\""
}
set defport [lindex $urlTypes($proto) 0]
set defcmd [lindex $urlTypes($proto) 1]
if {$port eq ""} {set port $defport}
if {![catch {$http(-proxyfilter) $host} proxy]} {
set phost [lindex $proxy 0]
set pport [lindex $proxy 1]
}
set url ${proto}://
if {$user ne ""} {
append url $user
append url @
}
append url $host
if {$port != $defport} {append url : $port}
append url $srvurl
set state(url) $url
set sockopts [list]
if {$state(-timeout) > 0} {
set state(after) [after $state(-timeout) [list http::reset $token timeout]]
lappend sockopts -async
}
if {[info exists phost] && ($phost ne "")} {
set srvurl $url
set targetAddr [list $phost $pport]
} else {set targetAddr [list $host $port]}
set state(socketinfo) $host:$port
if {$state(-keepalive)} {
variable socketmap
if {[info exists socketmap($state(socketinfo))]} {
if {[catch {fconfigure $socketmap($state(socketinfo))}]} {
Log "WARNING: socket for $state(socketinfo) was closed"
unset socketmap($state(socketinfo))
} else {
set sock $socketmap($state(socketinfo))
Log "reusing socket $sock for $state(socketinfo)"
catch {fileevent $sock writable {}}
catch {fileevent $sock readable {}}
}
}
set state(connection) {}
}
if {![info exists sock]} {
if {[info exists state(-myaddr)]} {lappend sockopts -myaddr $state(-myaddr)}
if {[catch {eval $defcmd $sockopts $targetAddr} sock]} {
set state(sock) $sock
Finish $token "" 1
cleanup $token
return -code error $sock
}
}
set state(sock) $sock
Log "Using $sock for $state(socketinfo)" [expr {$state(-keepalive)?"keepalive":""}]
if {$state(-keepalive)} {set socketmap($state(socketinfo)) $sock}
if {$state(-timeout) > 0} {
fileevent $sock writable [list http::Connect $token]
http::wait $token
if {![info exists state]} {
return $token
} elseif {$state(status) eq "error"} {
set err [lindex $state(error) 0]
cleanup $token
return -code error $err
} elseif {$state(status) ne "connect"} {return $token}
set state(status) ""
}
fconfigure $sock -translation {auto crlf} -buffersize $state(-blocksize)
catch {fconfigure $sock -blocking off}
set how GET
if {$isQuery} {
set state(querylength) [string length $state(-query)]
if {$state(querylength) > 0} {
set how POST
set contDone 0
} else {
unset state(-query)
set isQuery 0
}
} elseif {$state(-validate)} {set how HEAD} elseif {$isQueryChannel} {
set how POST
fconfigure $state(-querychannel) -blocking 1 -translation binary
set contDone 0
}
if {[info exists state(-method)] && $state(-method) ne ""} {set how $state(-method)}
if {[catch {
puts $sock "$how $srvurl HTTP/$state(-protocol)"
puts $sock "Accept: $http(-accept)"
array set hdrs $state(-headers)
if {[info exists hdrs(Host)]} {
puts $sock "Host: $hdrs(Host)"
} elseif {$port == $defport} {
puts $sock "Host: $host"
} else {puts $sock "Host: $host:$port"}
unset hdrs
puts $sock "User-Agent: $http(-useragent)"
if {$state(-protocol) == 1.0 && $state(-keepalive)} {puts $sock "Connection: keep-alive"}
if {$state(-protocol) > 1.0 && !$state(-keepalive)} {puts $sock "Connection: close"}
if {[info exists phost] && ($phost ne "") && $state(-keepalive)} {puts $sock "Proxy-Connection: Keep-Alive"}
set accept_encoding_seen 0
foreach {key value} $state(-headers) {
if {[string equal -nocase $key "host"]} {continue}
if {[string equal -nocase $key "accept-encoding"]} {set accept_encoding_seen 1}
set value [string map [list \n "" \r ""] $value]
set key [string trim $key]
if {[string equal -nocase $key "content-length"]} {
set contDone 1
set state(querylength) $value
}
if {[string length $key]} {puts $sock "$key: $value"}
}
if {!$accept_encoding_seen && ([package vsatisfies [package provide Tcl] 8.6] || [llength [package provide zlib]]) && !([info exists state(-channel)] || [info exists state(-handler)])} then {
puts $sock "Accept-Encoding: gzip, identity, *;q=0.1"
}
if {$isQueryChannel && $state(querylength) == 0} {
set start [tell $state(-querychannel)]
seek $state(-querychannel) 0 end
set state(querylength) [expr {[tell $state(-querychannel)] - $start}]
seek $state(-querychannel) $start
}
if {$isQuery || $isQueryChannel} {
puts $sock "Content-Type: $state(-type)"
if {!$contDone} {puts $sock "Content-Length: $state(querylength)"}
puts $sock ""
fconfigure $sock -translation {auto binary}
fileevent $sock writable [list http::Write $token]
} else {
puts $sock ""
flush $sock
fileevent $sock readable [list http::Event $sock $token]
}
if {![info exists state(-command)]} {
wait $token
if {$state(status) eq "error"} {return -code error [lindex $state(error) 0]}
}
} err]} then {
if {$state(status) ne "error"} {Finish $token $err 1}
cleanup $token
return -code error $err
}
return $token
}

proc http::data {token} {
variable $token
upvar 0 $token state
return $state(body)
}
proc http::status {token} {
if {![info exists $token]} {return "error"}
variable $token
upvar 0 $token state
return $state(status)
}
proc http::code {token} {
variable $token
upvar 0 $token state
return $state(http)
}
proc http::ncode {token} {
variable $token
upvar 0 $token state
if {[regexp {[0-9]{3}} $state(http) numeric_code]} {
return $numeric_code
} else {
return $state(http)
}
}
proc http::size {token} {
variable $token
upvar 0 $token state
return $state(currentsize)
}
proc http::meta {token} {
variable $token
upvar 0 $token state
return $state(meta)
}
proc http::error {token} {
variable $token
upvar 0 $token state
if {[info exists state(error)]} {
return $state(error)
}
return ""
}

# http::cleanup
#
#	Garbage collect the state associated with a transaction

proc http::cleanup {token} {
variable $token
upvar 0 $token state
if {[info exists state]} {
unset state
}
}

# http::Connect
#
#	This callback is made when an asyncronous connection completes.

proc http::Connect {token} {
variable $token
upvar 0 $token state
global errorInfo errorCode
if {[eof $state(sock)] || [string length [fconfigure $state(sock) -error]]} then {
Finish $token "connect failed [fconfigure $state(sock) -error]" 1
} else {
set state(status) connect
fileevent $state(sock) writable {}
}
return
}

# http::Write
#
#	Write POST query data to the socket

proc http::Write {token} {
variable $token
upvar 0 $token state
set sock $state(sock)
set done 0
if {[catch {
if {[info exists state(-query)]} {
puts -nonewline $sock \
[string range $state(-query) $state(queryoffset) \
[expr {$state(queryoffset) + $state(-queryblocksize) - 1}]]
incr state(queryoffset) $state(-queryblocksize)
if {$state(queryoffset) >= $state(querylength)} {
set state(queryoffset) $state(querylength)
set done 1
}
} else {
set outStr [read $state(-querychannel) $state(-queryblocksize)]
puts -nonewline $sock $outStr
incr state(queryoffset) [string length $outStr]
if {[eof $state(-querychannel)]} {
set done 1
}
}
} err]} then {
set state(posterror) $err
set done 1
}
if {$done} {
catch {flush $sock}
fileevent $sock writable {}
fileevent $sock readable [list http::Event $sock $token]
}
if {[string length $state(-queryprogress)]} {
eval $state(-queryprogress) [list $token $state(querylength) $state(queryoffset)]
}
}

# http::Event
#
#	Handle input on the socket

proc http::Event {sock token} {
variable $token
upvar 0 $token state
if {![info exists state]} {
Log "Event $sock with invalid token '$token' - remote close?"
if {![eof $sock]} {
if {[set d [read $sock]] ne ""} {Log "WARNING: additional data left on closed socket"}
}
CloseSocket $sock
return
}
if {$state(state) eq "connecting"} {
if {[catch {gets $sock state(http)} n]} {
return [Finish $token $n]
} elseif {$n >= 0} {set state(state) "header"}
} elseif {$state(state) eq "header"} {
if {[catch {gets $sock line} n]} {
return [Finish $token $n]
} elseif {$n == 0} {
if {$state(http) == "" || [lindex $state(http) 1] == 100} {
return
}
set state(state) body
if {$state(-validate)} {
Eof $token
return
}
if {!(([info exists state(connection)] && ($state(connection) eq "close")) || [info exists state(transfer)]) && ($state(totalsize) == 0)} then {
Log "body size is 0 and no events likely - complete."
Eof $token
return
}
fconfigure $sock -translation binary
if {$state(-binary) || ![string match -nocase text* $state(type)]} then {set state(binary) 1}
if {$state(binary) || [string match *gzip* $state(coding)] || [string match *compress* $state(coding)]} then {
if {[info exists state(-channel)]} {
fconfigure $state(-channel) -translation binary
}
}
if {[info exists state(-channel)] && ![info exists state(-handler)]} then {
fileevent $sock readable {}
CopyStart $sock $token
return
}
} elseif {$n > 0} {
if {[regexp -nocase {^([^:]+):(.+)$} $line x key value]} {
switch -- [string tolower $key] {
content-type {
set state(type) [string trim [string tolower $value]]
if {[regexp -nocase \
{charset\s*=\s*\"((?:[^""]|\\\")*)\"} \
$state(type) -> cs]} {
set state(charset) [string map {{\"} \"} $cs]
} else {
regexp -nocase {charset\s*=\s*(\S+?);?} \
$state(type) -> state(charset)
}
}
content-length {set state(totalsize) [string trim $value]}
content-encoding {set state(coding) [string trim $value]}
transfer-encoding {set state(transfer) [string trim [string tolower $value]]}
proxy-connection -
connection {set state(connection) [string trim [string tolower $value]]}
}
lappend state(meta) $key [string trim $value]
}
}
} else {
if {[catch {
if {[info exists state(-handler)]} {
set n [eval $state(-handler) [list $sock $token]]
} elseif {[info exists state(transfer_final)]} {
set line [getTextLine $sock]
set n [string length $line]
if {$n > 0} {
Log "found $n bytes following final chunk"
append state(transfer_final) $line
} else {
Log "final chunk part"
Eof $token
}
} elseif {[info exists state(transfer)] && $state(transfer) eq "chunked"} then {
set size 0
set chunk [getTextLine $sock]
set n [string length $chunk]
if {[string trim $chunk] ne ""} {
scan $chunk %x size
if {$size != 0} {
set bl [fconfigure $sock -blocking]
fconfigure $sock -blocking 1
set chunk [read $sock $size]
fconfigure $sock -blocking $bl
set n [string length $chunk]
if {$n >= 0} {
append state(body) $chunk
}
if {$size != [string length $chunk]} {
Log "WARNING: mis-sized chunk:\
was [string length $chunk], should be $size"
}
getTextLine $sock
} else {
set state(transfer_final) {}
}
}
} else {
set block [read $sock $state(-blocksize)]
set n [string length $block]
if {$n >= 0} {append state(body) $block}
}
if {[info exists state]} {
if {$n >= 0} {incr state(currentsize) $n}
if {($state(totalsize) > 0) && ($state(currentsize) >= $state(totalsize))} then {Eof $token}
}
} err]} then {
return [Finish $token $err]
} else {
if {[info exists state(-progress)]} {eval $state(-progress) [list $token $state(totalsize) $state(currentsize)]}
}
}
if {![catch {eof $sock} eof] && $eof} {
if {[info exists $token]} {
set state(connection) close
Eof $token
} else {
CloseSocket $sock
}
return
}
}

# http::getTextLine --
#
#	Get one line with the stream in blocking crlf mode

proc http::getTextLine {sock} {
set tr [fconfigure $sock -translation]
set bl [fconfigure $sock -blocking]
fconfigure $sock -translation crlf -blocking 1
set r [gets $sock]
fconfigure $sock -translation $tr -blocking $bl
return $r
}

# http::CopyStart
#
#	Error handling wrapper around fcopy

proc http::CopyStart {sock token} {
variable $token
upvar 0 $token state
if {[catch {
fcopy $sock $state(-channel) -size $state(-blocksize) -command [list http::CopyDone $token]
} err]} then {
Finish $token $err
}
}

# http::CopyDone
#
#	fcopy completion callback

proc http::CopyDone {token count {error {}}} {
variable $token
upvar 0 $token state
set sock $state(sock)
incr state(currentsize) $count
if {[info exists state(-progress)]} {eval $state(-progress) [list $token $state(totalsize) $state(currentsize)]}
if {[string length $error]} {
Finish $token $error
} elseif {[catch {eof $sock} iseof] || $iseof} {
Eof $token
} else {
CopyStart $sock $token
}
}

# http::Eof
#
#	Handle eof on the socket

proc http::Eof {token {force 0}} {
variable $token
upvar 0 $token state
if {$state(state) eq "header"} {
set state(status) eof
} else {set state(status) ok}
if {($state(coding) eq "gzip") && [string length $state(body)] > 0} {
if {[catch {
if {[package vsatisfies [package present Tcl] 8.6]} {
set state(body) [zlib gunzip $state(body)]
} else {
set state(body) [Gunzip $state(body)]
}
} err]} then {
return [Finish $token $err]
}
}
if {!$state(binary)} {
set enc [CharsetToEncoding $state(charset)]
if {$enc ne "binary"} {set state(body) [encoding convertfrom $enc $state(body)]}
set state(body) [string map {\r\n \n \r \n} $state(body)]
}
Finish $token
}

# http::wait --
#
#	See documentation for details.

proc http::wait {token} {
variable $token
upvar 0 $token state
if {![info exists state(status)] || $state(status) eq ""} {vwait ${token}(status)}
return [status $token]
}

# http::formatQuery --
#
#	See documentation for details.  Call http::formatQuery with an even
#	number of arguments, where the first is a name, the second is a value,
#	the third is another name, and so on.

proc http::formatQuery {args} {
set result ""
set sep ""
foreach i $args {
append result $sep [mapReply $i]
if {$sep eq "="} {
set sep &
} else {
set sep =
}
}
return $result
}

# http::mapReply --
#
#	Do x-www-urlencoded character mapping


proc http::mapReply {string} {
variable http
variable formMap
if {$http(-urlencoding) ne ""} {
set string [encoding convertto $http(-urlencoding) $string]
return [string map $formMap $string]
}
set converted [string map $formMap $string]
if {[string match "*\[\u0100-\uffff\]*" $converted]} {
regexp {[\u0100-\uffff]} $converted badChar
return -code error "can't read \"formMap($badChar)\": no such element in array"
}
return $converted
}

# http::ProxyRequired --
#
#	Default proxy filter.

proc http::ProxyRequired {host} {
variable http
if {[info exists http(-proxyhost)] && [string length $http(-proxyhost)]} {
if {![info exists http(-proxyport)] || ![string length $http(-proxyport)]} then {set http(-proxyport) 8080}
return [list $http(-proxyhost) $http(-proxyport)]
}
}

# http::CharsetToEncoding --
#
#	Tries to map a given IANA charset to a tcl encoding.  If no encoding
#	can be found, returns binary.
#

proc http::CharsetToEncoding {charset} {
variable encodings
set charset [string tolower $charset]
if {[regexp {iso-?8859-([0-9]+)} $charset -> num]} {
set encoding "iso8859-$num"
} elseif {[regexp {iso-?2022-(jp|kr)} $charset -> ext]} {
set encoding "iso2022-$ext"
} elseif {[regexp {shift[-_]?js} $charset]} {
set encoding "shiftjis"
} elseif {[regexp {(?:windows|cp)-?([0-9]+)} $charset -> num]} {
set encoding "cp$num"
} elseif {$charset eq "us-ascii"} {
set encoding "ascii"
} elseif {[regexp {(?:iso-?)?lat(?:in)?-?([0-9]+)} $charset -> num]} {
switch -- $num {
5 {set encoding "iso8859-9"}
1 - 2 - 3 {
set encoding "iso8859-$num"
}
}
} else {
set encoding $charset
}
set idx [lsearch -exact $encodings $encoding]
if {$idx >= 0} {
return $encoding
} else {
return "binary"
}
}

# http::Gunzip --
#
#	Decompress data transmitted using the gzip transfer coding.
#

proc http::Gunzip {data} {
binary scan $data Scb5icc magic method flags time xfl os
set pos 10
if {$magic != 0x1f8b} {return -code error "invalid data: supplied data is not in gzip format"}
if {$method != 8} {return -code error "invalid compression method"}
foreach {f_text f_crc f_extra f_name f_comment} [split $flags ""] break
set extra ""
if {$f_extra} {
binary scan $data @${pos}S xlen
incr pos 2
set extra [string range $data $pos $xlen]
set pos [incr xlen]
}
set name ""
if {$f_name} {
set ndx [string first \0 $data $pos]
set name [string range $data $pos $ndx]
set pos [incr ndx]
}
set comment ""
if {$f_comment} {
set ndx [string first \0 $data $pos]
set comment [string range $data $pos $ndx]
set pos [incr ndx]
}
set fcrc ""
if {$f_crc} {
set fcrc [string range $data $pos [incr pos]]
incr pos
}
binary scan [string range $data end-7 end] ii crc size
set inflated [zlib inflate [string range $data $pos end-8]]
set chk [zlib crc32 $inflated]
if {($crc & 0xffffffff) != ($chk & 0xffffffff)} {return -code error "invalid data: checksum mismatch $crc != $chk"}
return $inflated
}